//
//   @file    amcfastps_interface.cxx         
//   
//
//   @author M.Sutton
// 
//   Copyright (C) 2013 M.Sutton (sutt@cern.ch)    
//
//   $Id: amcfastps_interface.cxx, v0.0   Tue 13 Aug 2013 12:20:03 CEST sutt $




#include <cstdlib>
#include <iostream>
#include <cmath>

// applgrid header files
#include "appl_grid/appl_grid.h"
#include "appl_grid/lumi_pdf.h"

// LHAPDF
#include "LHAPDF/LHAPDF.h"

// amcfast header files
#include "amcfastps_interface.h"
#include "amcfast_utils.h"
#include "amcatnlo_common.h"

// ROOT header files
#include "TPad.h"

/// lhapdf functions
extern "C" double alphasPDF_(const double& );
extern "C" void evolvePDF_(const double&,const double&, double* );


////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////// NLO+PS mode of aMCatNLO ////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

 
/// extern hook  functions for amc@nlo control 
extern void (*applmc_initptr)();
extern void (*applmc_fillrefptr)();
extern void (*applmc_termptr)();


///  calling this function sets up the hooks for the applgrid control
///  functions - without this function beging called, no applgrid code
///  will be executed

extern "C" bool setup_amcfastmc_() {

  std::cerr << "setup_amcatnlomc() setting up function hooks" << std::endl;
 
  applmc_initptr    = amcfastmc::init;
  applmc_fillrefptr = amcfastmc::fill_ref;
  applmc_termptr    = amcfastmc::term;

  return true;
}


namespace amcfastmc {  


bool amcfastmc_ready = setup_amcfastmc_();
    


///
///  set up the grid, create the parton luminosities
///

// To begin with we hardcode two grids
// corresponding to two different observables
appl::grid* grid_obsA = 0;
appl::grid* grid_obsB = 0;


// Different names for reading and writing of grids
std::string grid_filenameA_in  = "grid_obsA_in.root";
std::string grid_filenameA_out = "grid_obsA_out.root";

std::string grid_filenameB_in = "grid_obsB_in.root";
std::string grid_filenameB_out = "grid_obsB_out.root";

// Test root histograms
//TH1D* grid_testA = 0;
//TH1D* grid_testB = 0;

///
/// Initialization
///

void init() { 
    
    std::cerr << "C++, mc, init()" << std::endl;
    
    // --------- Some aMCatNLO initialization
    //leading_order -> Power of alphas of Born events
    //int leading_order =  appl_common_fixed_.bpower;
    int leading_order = 0; // Now hardwired, to get from common block

    // Number of loops 
    // Actually this is the number of independent grids in aMCatNLO
    // At the NLO+PS mode there are two grids to fill,
    // corresponding to the full NLO+PS calculation
    // One for Born like kinematics and the other for resolved kinematics
    int nloops=1; // Single grid
    
    // -------- Define the settings for the interpolation in x and Q2
    // -------- These are commons to all the grids computed

    // Number of points in Q2 interpolation
    int NQ2=20;
    
    // Max and min value of Q2
    double Q2min=100.0;
    double Q2max=500000.0;
    
    // Order of polynomial interpolation in Q2
    int Q2order=5;
    
    // Number of points in x interpolation
    int Nx=60;
    
    // Min and max value of x
    double xmin=1e-5;
    double xmax=0.9;
    
    // Order of polynomial interpolation in x
    int xorder=5;

    // Read PDF luminosities from file
    ifstream lumifile;
    std::cerr<<"Reading info from initial_states_map.dat"<<std::endl;
    lumifile.open("initial_states_map.dat");
    if(lumifile.fail()){
      std::cerr<<"Unable to open initial_states_map.dat"<<std::endl;exit(-10);
    }
    std::string line1;
    int applmc_nlumi=0;
    int applmc_nproc[121]={0};
    int applmc_lumimap[121][121][2]={{{0}}};

    while(getline(lumifile,line1)){
      std::istringstream linestream(line1);
      int applmc_ilumi=0, nproc=0;
      linestream>>applmc_ilumi>>nproc;
      //    std::cout<<applmc_ilumi<<" "<<nproc<<std::endl;
      applmc_nproc[applmc_ilumi-1]=nproc;
      std::cout<<applmc_ilumi<<" "<<applmc_nproc[applmc_ilumi-1]<<std::endl;
      // Check
      if(applmc_ilumi<=0 || applmc_ilumi>121 ){
	std::cerr<<"Invalid value of applmc_ilumi"<<std::endl;
	std::cerr<<applmc_ilumi<<std::endl;exit(-10);
      }
      if(applmc_nproc[applmc_ilumi-1]<=0 || applmc_nproc[applmc_ilumi-1]>121 ){
	std::cerr<<"Invalid value of applmc_nproc[applmc_ilumi-1]"<<std::endl;
	std::cerr<<applmc_nproc[applmc_ilumi-1]<<std::endl;exit(-10);
      }
      // std::cout<<applmc_ilumi<<" "<<applmc_nproc[applmc_ilumi-1]<<std::endl;
      for(int iproc=0;iproc<applmc_nproc[applmc_ilumi-1];iproc++){
	linestream>>applmc_lumimap[applmc_ilumi-1][iproc][0]>>
	  applmc_lumimap[applmc_ilumi-1][iproc][1];
	// Check
	// Use LHAPDF notation
	if(applmc_lumimap[applmc_ilumi-1][iproc][0]==21)applmc_lumimap[applmc_ilumi-1][iproc][0]=0;
	if(applmc_lumimap[applmc_ilumi-1][iproc][1]==21)applmc_lumimap[applmc_ilumi-1][iproc][1]=0;
	int pdg1 = applmc_lumimap[applmc_ilumi-1][iproc][0];
	int pdg2 = applmc_lumimap[applmc_ilumi-1][iproc][1];
	if( pdg1<-5|| pdg1>5 ||  pdg2<-5|| pdg2>5 ){
	  std::cerr<<"Invalied values of PDG IDs "<<std::cerr<<std::endl;
	  std::cerr<<pdg1<<" "<<pdg2<<std::endl;std::exit(-10);
	}
	std::cout<<"\t"<<applmc_lumimap[applmc_ilumi-1][iproc][0]
		 <<"\t"<<applmc_lumimap[applmc_ilumi-1][iproc][1]<<std::endl;
      }
      // Set the value of the total number of luminosities
      applmc_nlumi = applmc_ilumi;
    }
    lumifile.close();
    std::cerr<<"\n nlumi = "<<applmc_nlumi<<std::endl;
    // Check reasonable value of luminosities
    if(applmc_nlumi <= 0 || applmc_nlumi >121){
      std::cerr<<"Invalid value of nlumi"<<std::endl;exit(-10);
    }
    
        // ---- PDF luminosities ---------
    // set up the applgrid pdf luminosities
    std::vector<int> pdf_luminosities;
    pdf_luminosities.push_back( applmc_nlumi );
    
    /// loop over different parton luminosities
    for ( int ilumi=0 ; ilumi< applmc_nlumi ; ilumi++ ){
      int nproc = applmc_nproc[ilumi];
      
      pdf_luminosities.push_back( ilumi );
      pdf_luminosities.push_back( nproc );
      
      /// loop over different parton-parton combinations within each luminosity
      for ( int iproc=0 ; iproc<nproc ; iproc++ ){
	pdf_luminosities.push_back( applmc_lumimap[ilumi][iproc][0] );
	pdf_luminosities.push_back( applmc_lumimap[ilumi][iproc][1] );
      }
    }
   
    /// ensure that the pdf combination routine is set up correctly - 
    /// use a name for the pdf combination type ending with .config 
    /// - this will configure from a file with the same format as the 
    /// amcatnlo initial_states_map.dat file usless a serialised 
    /// vector including the combination dete is also passed into the 
    /// constructor - as is the case here
    std::string genpdf="amcatnlo.config";
    
    /// don't need to save the pointer - automatically gets 
    /// added to the factory    
    ///  bool amcflag = true; /// flag that this is an amcatnlo process
    new lumi_pdf( genpdf, pdf_luminosities ); //, amcflag ); 
   
    // Check that info is being properly passed
    std::cout<<"\n obsA width, min, max, type = "<<
      applmc_common_histokina_.obsA_binwidth<<" "<<
      applmc_common_histokina_.obsA_min<<" "<<
      applmc_common_histokina_.obsA_max<<" "<<
      applmc_common_histokina_.obsA_type<<" "<<std::endl;
    std::cout<<"\n obsB width, min, max, type = "<<
      applmc_common_histokinb_.obsB_binwidth<<" "<<
      applmc_common_histokinb_.obsB_min<<" "<<
      applmc_common_histokinb_.obsB_max<<" "<<
      applmc_common_histokinb_.obsB_type<<" "<<std::endl;
    // Some check
    if( applmc_common_histokina_.obsA_binwidth==0.0 ||
	applmc_common_histokinb_.obsB_binwidth==0.0 ){
      std::cerr<<"Invalid value of histo bin width "<<std::endl;
      exit(-10);
    }

    // Check that the observable is among the list of 
    // possible observables
    // 1 -> y3 
    // 2 -> pt3
    // 3 -> y34 
    // 4 -> m34
    if( applmc_common_histokina_.obsA_type < 1 || 
	applmc_common_histokina_.obsA_type > 4){
      std::cout<<"Invalid value for obsA_type = "<<applmc_common_histokina_.obsA_type<<std::endl;
      exit(-10);
    }
    if( applmc_common_histokinb_.obsB_type < 1 || 
	applmc_common_histokinb_.obsB_type > 4){
      std::cout<<"Invalid value for obsB_type = "<<applmc_common_histokinb_.obsB_type<<std::endl;
      exit(-10);
    }

    // ------ Grid Initialization -----------------
    // ------ Definition of the observables ------------
    // First of all we need to define the physical observables
    // We need one applgrid for each observable
    // For the time being we define a single observable, 
    // the rapidity distribution of charged leptons in  p p > l+ nl,
    // The default is an histogram with equally space widths
    // Changing histo widths can be done using the standard ROOT modules
    // Note that these definitions should match those of madfks_plot.f
    
    /// does the grid file exist? If so read the grid from the file, 
    /// otherwise create a new grid from scratch 
    
    // Observable A
    if ( amcfast::file_exists( grid_filenameA_in ) ) { 
      grid_obsA = new appl::grid( grid_filenameA_in );
      grid_obsA->optimise(true);

      // Create test root histogram
      //int Nobs= int( (applmc_common_histokina_.obsA_max - applmc_common_histokina_.obsA_min ) / applmc_common_histokina_.obsA_binwidth +1e-3);
      //double obsmin=applmc_common_histokina_.obsA_min;
      //double obsmax=applmc_common_histokina_.obsA_max; 
      //if(grid_testA==0)grid_testA = new TH1D("grid_testA","grid_testA",Nobs,obsmin,obsmax);
      
    }
    else { 
      
      int Nobs= int( (applmc_common_histokina_.obsA_max - applmc_common_histokina_.obsA_min ) / applmc_common_histokina_.obsA_binwidth  +1e-4);
      double obsmin=applmc_common_histokina_.obsA_min;
      double obsmax=applmc_common_histokina_.obsA_max; 
      
      // Grid initialization
      grid_obsA = new appl::grid(NQ2,  Q2min ,Q2max,  Q2order,  
				 Nx,  xmin,    xmax,         xorder,
				 Nobs, obsmin,  obsmax, 
				 genpdf, 
				 leading_order, nloops);
      
      /// tell the grid it needs the amcatnlo convolution
      grid_obsA->amcatnlo();

      // Check number of entries
      std::cout<<"Number of entries in grid_obsA ref histo = "
	       <<grid_obsA->getReference()->GetEntries()<<std::endl;
      
      // Create test root histogram
      //if(grid_testA==0)grid_testA = new TH1D("grid_testA","grid_testA",Nobs,obsmin,obsmax);
      
    }

    // Observable B
    if ( amcfast::file_exists( grid_filenameB_in ) ) { 
      grid_obsB = new appl::grid( grid_filenameB_in );
      grid_obsB->optimise(true);
      
    }
    else { 
      
      int Nobs= int( (applmc_common_histokinb_.obsB_max - applmc_common_histokinb_.obsB_min ) / applmc_common_histokinb_.obsB_binwidth +1e-3);
      double obsmin=applmc_common_histokinb_.obsB_min;
      double obsmax=applmc_common_histokinb_.obsB_max; 
      
      // Grid initialization
      grid_obsB = new appl::grid(NQ2,  Q2min ,Q2max,  Q2order,  
				 Nx,  xmin,    xmax,         xorder,
				 Nobs, obsmin,  obsmax, 
				 genpdf, 
				 leading_order, nloops);

      /// tell the grid it needs the amcatnlo convolution
      grid_obsB->amcatnlo();

    }

    // Repeat if other distributions are to be interpolated
    
  // ------------ applgrid initialization completed
    
    std::cerr << "exit C++, mc, init()" << std::endl;
    
}





///
/// Physical observables
/// 
void getObservables(double & obsA, double & obsB){

    // Initialization
    int const ipart = 0;
  
    // Kinematics
    double E=applmc_common_histofill_.pplab[ipart][0];
    double px=applmc_common_histofill_.pplab[ipart][1];
    double py=applmc_common_histofill_.pplab[ipart][2];
    double pz=applmc_common_histofill_.pplab[ipart][3]; 

    // Physical observables
    obsA =0.0;
    obsB=0.0;

    // 1 -> y3 
    // 2 -> pt3
    // 3 -> y34 
    // 4 -> m34
    if(  applmc_common_histokina_.obsA_type == 1 ){
      // Rapidity of particle 3 (in aMCatNLO record)
      obsA = amcfast::rapidity(E,pz);
    }
    else if(  applmc_common_histokina_.obsA_type == 3 ){
      // Rapidity of the sum of particles 3 and 4 (in aMCatNLO record)
      // The sum has been done already at the level of HWANAL
      obsA = amcfast::rapidity(E,pz);
    }
    else{
      std::cout<<"Invalid value of obsA_type = "
	       <<applmc_common_histokina_.obsA_type<<std::endl;
      exit(-10);
    }

    if(  applmc_common_histokinb_.obsB_type == 2 ){
      // pt of particle 3 (in aMCatNLO record)
      obsB = amcfast::pt ( px,  py );   
    }
    else if(  applmc_common_histokinb_.obsB_type == 4 ){
      // Invariant mass of the sum of particles 3 and 4 (in aMCatNLO record)
      // The sum has been done already at the level of HWANAL
      obsB = amcfast::mass(E,px,py,pz);
    }
    else{
      std::cout<<"Invalid value of obsB_type = "
	       <<applmc_common_histokinb_.obsB_type<<std::endl;
      exit(-10);
    }

    // Check that fortran and C++ are identical
    double const tol=1e-4;
    double obsA_f=applmc_common_histofill_.kinhistoA;
    double obsB_f=applmc_common_histofill_.kinhistoB;
    double diffA=std::fabs( ( obsA - obsA_f ) / obsA_f );
    double diffB=std::fabs( ( obsB - obsB_f ) / obsB_f );
    if( std::fabs(obsA_f) > 1e-20 && diffA > tol ){
      std::cout<<"Mismatch in obsA"<<std::endl;
      std::cout<<"C++, obsA = "<<" "<<obsA<<std::endl;
      std::cout<<"F, obsA = "<<" "<<obsA_f<<std::endl;
      exit(-10);
    }
    if( std::fabs(obsB_f) > 1e-20 && diffB > tol ){
      std::cout<<"Mismatch in obsB"<<std::endl;
      std::cout<<"C++, obsB = "<<" "<<obsB<<std::endl;
      std::cout<<"F, obsB = "<<" "<<obsB_f<<std::endl;
      exit(-10);
    }

    return;
    
}






///
///  Fill the applgrid reference histogram
///  Analog of  HWANAL in the herwig analysis routine
///

void fill_ref() { 
    
    std::cerr << "\n C++, mc, in fill_ref()" << std::endl;
    
    // Get the reference weight
    double www = applmc_common_histofill_.www_histo;
    std::cerr<<"Reference weight = "<<www<<std::endl;
    
    // Get the physical observables
    double obsA=0, obsB=0;
    getObservables(obsA, obsB);

    std::cerr<<"obsA, obsB = "<<obsA<<" "<<obsB<<std::endl;

    // Reference histogram
    static int ic=0; // Sets up counter
    std::cout<<"C++ mc -> ic, obsA, www_ref = "<<++ic<<" "<<obsA<<" "<<www<<std::endl;
    grid_obsA->getReference()->Fill(obsA,www);
    grid_obsB->getReference()->Fill(obsB,www);
    std::cout<<"Number of entries of C++ ref histo = "<< grid_obsA->getReference()->GetEntries()<<std::endl;
   
    // Fill also test root histograms
    //if(grid_testA) grid_testA->Fill(obsA,www);
        
    // Done
    std::cerr << "C++, mc, exit fill_ref()" << std::endl;
    
}





///
/// comparison of histograms
///
// Compare histograms
bool CompareHisto(std::string Obs){
  
    std::cout<<"In compareHisto, Obs = "<<Obs<<std::endl;

    std::string const ObsA="A";
    std::string const ObsB="B";
    if(Obs!=ObsA && Obs!=ObsB ){
      std::cout<<"Invalid value for Obs = "<<Obs<<std::endl;
      exit(-10);
    }

    // Get number of bins of the histogram
    int Nobs= 1;
    if(Obs==ObsA) Nobs = int( (applmc_common_histokina_.obsA_max - applmc_common_histokina_.obsA_min ) / applmc_common_histokina_.obsA_binwidth +1e-3);
    if(Obs==ObsB) Nobs = int( (applmc_common_histokinb_.obsB_max - applmc_common_histokinb_.obsB_min ) / applmc_common_histokinb_.obsB_binwidth +1e-3);
    std::cout<<"Nobs = "<<Nobs<<std::endl;

    // Initialization
    double amcref_kin[1000]={0.0}, amcref_xsec[1000]={0.0};
    
    ifstream madtd;
    std::string filename;
    if(Obs == ObsA) filename="aMCatNLO_obsA.top";
    if(Obs == ObsB) filename="aMCatNLO_obsB.top";

    std::cout<<"Open file = "<<filename<<std::endl;
    madtd.open(filename.c_str()); 
    if(madtd.fail()){
      std::cout<<"Unable to open "<<filename<<std::endl;
      exit(-10);
    }

    std::string line1;
    bool gl=false;
    for(int i=0;i<12;i++){
      gl = getline(madtd,line1);
    }
    for(int i=0;i<Nobs;i++){
      gl = getline(madtd,line1);
      std::istringstream linestream(line1);
      linestream.precision(20);
      linestream>>amcref_kin[i]>>amcref_xsec[i];
      std::cout<<amcref_kin[i]<<"  "<<amcref_xsec[i]<<std::endl;
    }
    madtd.close();
      
    std::cout<<"Obs =  "<<Obs<<std::endl;
    std::cout<<"ibin, RefHisto,  aMCatNLO "<<std::endl;
    // Check that MADatNLO and RefHistoA output are identical
    double const tol=1e-2;
    // open unit to save data
    std::ofstream outobs;

    if(Obs==ObsA){
      outobs.open("HistoComp_obsA.res");
      for(int ibin=1;ibin<=grid_obsA->getReference()->GetNbinsX();ibin++){
	//	std::cout<<ibin<<"   "<< grid_obsA->getReference()->GetBinContent(ibin)<<"   "<<"   "<<amcref_xsec[ibin-1]<<std::endl;
	outobs<<amcref_kin[ibin-1] <<"   "<< grid_obsA->getReference()->GetBinContent(ibin)<<"   "
		 <<"   "<<amcref_xsec[ibin-1]<<std::endl;
	if( fabs( amcref_xsec[ibin-1] ) > 1e-20 ){
	  double diff=fabs( (grid_obsA->getReference()->GetBinContent(ibin) -amcref_xsec[ibin-1] ) / amcref_xsec[ibin-1] );
	  if(diff>tol){std::cerr<<"RefHisto and aMCatNLO not identical!"<<std::endl;exit(-10);}
	}
      }
    }  
    if(Obs==ObsB){
      outobs.open("HistoComp_obsB.res");
      for(int ibin=1;ibin<=grid_obsB->getReference()->GetNbinsX();ibin++){
	std::cout<<ibin<<"   "<< grid_obsB->getReference()->GetBinContent(ibin)<<"   " <<"   "<<amcref_xsec[ibin-1]<<std::endl;
	outobs<<amcref_kin[ibin-1] <<"   "<< grid_obsB->getReference()->GetBinContent(ibin)<<"   "
		 <<"   "<<amcref_xsec[ibin-1]<<std::endl;
	if( fabs( amcref_xsec[ibin-1] ) > 1e-20 ){
	  double diff=fabs( (grid_obsB->getReference()->GetBinContent(ibin) -amcref_xsec[ibin-1] ) / amcref_xsec[ibin-1] );
	  if(diff>tol){std::cerr<<"RefHisto and aMCatNLO not identical!"<<std::endl;exit(-10);}
	}
      }
    }  
    outobs.close();
    
    return true;

}






///
///  Finish up for the applgrid grid structure, write out the grids
///  Also apply the same normalization factor as in  HWAEND
///  in order to match the aMCatNLO histograms
///

void term() { 
    
    std::cerr << "C++, mc, term()" << std::endl;

    // Final normalization factor
    double norm = applmc_common_histonorm_.histonorm;
    // Check reasonable value
    if(norm <= 0.0 || norm > 1e50){
      std::cerr<<"Invalid value for normalization = "<<norm<<std::endl;
      exit(-10);
    }

    // Write out all the grids that have been filled
    // Save only if the pointer has been initialized
    if(grid_obsA) {

      // Set correct normalization
      grid_obsA->getReference()->Scale(norm);
          
      // Compare histogram with original aMCatNLO
      std::string Obs="A";
      bool check = CompareHisto(Obs);
      if(!check) exit(-10);
      
      grid_obsA->run() = 1; // set the number of runs to 1
      // Write the grid to file
      grid_obsA->Write(grid_filenameA_out);
      
    }

    if(grid_obsB) {

      // Set correct normalization
      grid_obsB->getReference()->Scale(norm);
          
      // Compare histogram with original aMCatNLO
      std::string Obs="B";
      bool check = CompareHisto(Obs);
      if(!check) exit(-10);

      grid_obsB->run() = 1; // set the number of runs to 1
      // Write the grid to file
      grid_obsB->Write(grid_filenameB_out);
      
    }
    
    std::cerr << "C++, mc, exit term()" << std::endl;
    
    // Then we need to combine the various grids, using exactly
    // the same code as the combine_events routine in aMCatNLO
    // This is done by the applgrid-combination script
    
  }
    
}  // end namespace amcfastmc

